///*|-----------------------------------------------------------------------------
// *|            This source code is provided under the Apache 2.0 license      --
// *|  and is provided AS IS with no warranty or guarantee of fit for purpose.  --
// *|                See the project's LICENSE.md for details.                  --
// *|             Copyright Refinitiv 2021. All rights reserved.                --
///*|-----------------------------------------------------------------------------

#include <cstdlib>
#include <stdexcept>
#include <cstring>
#include "MRNConsumer.h"
#include "zlib.h"

using namespace refinitiv::ema::access;
using namespace refinitiv::ema::rdm;
using namespace std;

// RefreshMsg contains feed specific metadata - will not contain NTA fragment 
void AppClient::onRefreshMsg( const RefreshMsg& refreshMsg, const OmmConsumerEvent& ) 
{
	if ( refreshMsg.hasMsgKey() )
		cout << endl << "REFRESH Item Name: " << refreshMsg.getName() << endl << "Service Name: " << refreshMsg.getServiceName();

	cout << endl << "Item State: " << refreshMsg.getState().toString() << endl;

	// Confirm the payload and domain type and dump the metadata Fields to the console
	if (DataType::FieldListEnum == refreshMsg.getPayload().getDataType() && (MMT_NEWS_TEXT_ANALYTICS == refreshMsg.getDomainType()))
		decode(refreshMsg.getPayload().getFieldList());
	else
		cerr << endl << "Wrong Domain and/or Payload " << refreshMsg.getDomainType() << " " << refreshMsg.getPayload().getDataType() << endl;
}

// UpdateMsg contains NTA fragment and item related metadata 
void AppClient::onUpdateMsg( const UpdateMsg& updateMsg, const OmmConsumerEvent& ) 
{
	if ( updateMsg.hasMsgKey() )
		cout << endl << "UPDATE Item Name: " << updateMsg.getName() << endl << "Service Name: " << updateMsg.getServiceName() << endl;

	// Confirm the payload and domain type is as expected and then attempt to extract the NTA Item 
	if (DataType::FieldListEnum == updateMsg.getPayload().getDataType() && (MMT_NEWS_TEXT_ANALYTICS == updateMsg.getDomainType()))
		decodeNTA(updateMsg.getPayload().getFieldList());
	else
		cerr << endl << "Wrong Domain and/or Payload " << updateMsg.getDomainType() << " " << updateMsg.getPayload().getDataType() << endl;
}

void AppClient::onStatusMsg( const StatusMsg& statusMsg, const OmmConsumerEvent& ) 
{
	if ( statusMsg.hasMsgKey() )
		cout << endl << "Item Name: " << statusMsg.getName() << endl << "Service Name: " << statusMsg.getServiceName();

	// Something not right? Report the status
	if (statusMsg.hasState())
		cout << endl << "Item State: " << statusMsg.getState().toString() << endl;
}

// Extract and dump the fields to the screen (excluding the zipped FRAGMENT field)
void AppClient::decode( const FieldList& fl )
{
	while (fl.forth())
	{
		const FieldEntry& fe = fl.getEntry();

		// Display FID, Field Name, Field Data type and actual value to screen
		cout << "Fid: " << fe.getFieldId() << " Name = " << fe.getName() << " DataType: " << DataType(fe.getLoad().getDataType()) << " Value: ";
		cout << "Fid: " << fe.getFieldId() << " Name = " << fe.getName() << " DataType: " << DataType( fe.getLoad().getDataType() ) << " Value: ";

		if (fe.getCode() == Data::BlankEnum)
			cout << "<BLANK>" << endl;
		else
			switch (fe.getLoadType())
		{
			case DataType::RealEnum:
				cout << fe.getReal().getAsDouble() << endl;
				break;
			case DataType::DateEnum:
				cout << (UInt64)fe.getDate().getDay() << " / " << (UInt64)fe.getDate().getMonth() << " / " << (UInt64)fe.getDate().getYear() << endl;
				break;
			case DataType::TimeEnum:
				cout << (UInt64)fe.getTime().getHour() << ":" << (UInt64)fe.getTime().getMinute() << ":" << (UInt64)fe.getTime().getSecond() << ":" << (UInt64)fe.getTime().getMillisecond() << endl;
				break;
			case DataType::IntEnum:
				cout << fe.getInt() << endl;
				break;
			case DataType::UIntEnum:
				cout << fe.getUInt() << endl;
				break;
			case DataType::AsciiEnum:
				cout << fe.getAscii() << endl;
				break;
			case DataType::RmtesEnum:
				cout << fe.getRmtes().toString() << endl;
				break;
			case DataType::EnumEnum:
				cout << fe.getEnum() << endl;
				break;
			case DataType::ErrorEnum:
				cout << "( " << fe.getError().getErrorCodeAsString() << " )" << endl;
				break;
			case DataType::BufferEnum:			// Assuming this is the FRAGMENT field - we are not expecting any other Buffer type fields
				cout << "<BUFFER>" << endl;		// Don't bother to try and display FRAGMENT as it is zipped and would appear as gobbledygook
				break;
			default:
				cout << endl;
				break;
		}
	}
}

// Handle an Update message which contains a NTA Item Fragment
void AppClient::decodeNTA(const FieldList& envelope)
{
	cout << "DecodeNTA" << endl;

	// Dump the FieldList first for informational purposes
	decode(envelope);

	NTAItem item;
	// Now lets extract the relevant fields into a NTA Item - including the Fragment
	item.init(envelope);

	// Is this 1st fragment of a NTA Item?
	if (item.getFragNum() == 1)
	{
		// Is it the complete NTA Item ?
		if (item.isComplete())	// i.e. check length of this single fragment buffer == expected total size of NTA Item
		{
			std::cout << "<<Single Fragment >>" << endl;
			string unzipped;
			unzip(item.getBuffer(), unzipped);	// decompress the buffer
			std::cout << unzipped << endl << endl;	// Display the decompressed NTA Item
			return;
		}
		else
		{							// Start of a multi-fragment item
			_multiFragItem = item;	// so copy this first fragment into our pending item instance
		}
			
	}
	else
	{	// FRAG_NUM > 1 so continue to build up NTA Item
		_multiFragItem.addFragment(item);	// Add the newly received fragment to the pending instance
		if (_multiFragItem.isComplete())	// Does length of the buffer now == expected total size of NTA Item ?
		{
			std::cout << "<<Multi Fragment complete>>" << endl;
			string unzipped;
			unzip(_multiFragItem.getBuffer(), unzipped);	// Decompress the complete buffer
			std::cout << unzipped << endl << endl;			// Display the decompressed NTA Item
			return;
		}
	}
}

void AppClient::unzip(const std::string& zipped, std::string& unzipped)
{
	if (zipped.size() == 0) {
		unzipped = zipped;
	}

	unzipped.clear();

	unsigned unzipBufferLength = zipped.size() * 4; // create a buffer that is about large enough to hold the output
	unsigned additionalUnzipBufferLength = unzipBufferLength;

	char* unzipBuffer = (char*)calloc(sizeof(char), unzipBufferLength);

	z_stream zStrm;
	zStrm.zalloc = Z_NULL;
	zStrm.zfree = Z_NULL;
	zStrm.next_in = (Bytef*)zipped.c_str();
	zStrm.avail_in = zipped.size();
	zStrm.total_out = 0;

	bool finished = false;

	if (inflateInit2(&zStrm, (16 + MAX_WBITS)) != Z_OK) {
		free(unzipBuffer);
		throw std::runtime_error("Unable to unzip data: inflateInit2 failed");
	}

	while (!finished) {
		// increase buffer size if it's too small
		if (zStrm.total_out >= unzipBufferLength) {
			char* unzipBuffer2 = (char*)calloc(sizeof(char), unzipBufferLength + additionalUnzipBufferLength);
			memcpy(unzipBuffer2, unzipBuffer, unzipBufferLength);
			unzipBufferLength += additionalUnzipBufferLength;
			free(unzipBuffer);
			unzipBuffer = unzipBuffer2;
		}

		zStrm.next_out = (Bytef *)(unzipBuffer + zStrm.total_out);
		zStrm.avail_out = unzipBufferLength - zStrm.total_out;

		int error = inflate(&zStrm, Z_SYNC_FLUSH);
		if (error == Z_STREAM_END) {
			finished = true;
		}
		else if (error != Z_OK)  {
			break;
		}
	}

	if (inflateEnd(&zStrm) != Z_OK) {
		free(unzipBuffer);
		throw std::runtime_error("Unable to unzip data: inflateEnd failed");
	}

	unzipped.reserve(zStrm.total_out);
	unzipped.append(unzipBuffer, zStrm.total_out);

	free(unzipBuffer);
}


int main( int argc, char* argv[] )
{ 
	try { 

		AppClient client;
		// Create consumer and specify our connectivity parameters + user name
		OmmConsumer consumer( OmmConsumerConfig().host( "ADS_TOR:14002" ).username( "user" ) );
		// Request NTA domain MRN_STORY type items 
		consumer.registerClient(ReqMsg().serviceName("ELEKTRON_DD").name("MRN_STORY").domainType(MMT_NEWS_TEXT_ANALYTICS), client);
		// Sleep long enough to receive a few stories
		sleep( 60000 );			// API calls onRefreshMsg(), onUpdateMsg() and onStatusMsg()
		
	} catch ( const OmmException& excp ) {
		cout << excp << endl;
	}
	return 0;
}
