///*|-----------------------------------------------------------------------------
// *|            This source code is provided under the Apache 2.0 license      --
// *|  and is provided AS IS with no warranty or guarantee of fit for purpose.  --
// *|                See the project's LICENSE.md for details.                  --
// *|             Copyright Refinitiv 2021. All rights reserved.                --
///*|-----------------------------------------------------------------------------

#include "Step4.h"

using namespace refinitiv::ema::access;
using namespace refinitiv::ema::rdm;
using namespace std;


void AppClient::onRefreshMsg( const RefreshMsg& refreshMsg, const OmmConsumerEvent& event) {
	// an item handle uniquely identifies the request and is used to unsubscribe it
	// a closure is any user owned object passed on when making a request and is passed back in response (refresh/update/status) messages
	cout << "Refresh message, item Handle: " << event.getHandle() << " Closure: " << event.getClosure() << endl;
	
	//  display item and service name
	if( refreshMsg.hasMsgKey() )
		cout << endl << "Item Name: " << refreshMsg.getName() << " (Service Name: " << refreshMsg.getServiceName() << ")";

	//  display the status of current request
	cout << endl << "Item State: " << refreshMsg.getState().toString() << endl;

	// Level2 data payload is encoded as an OMM Map or NoData if no payload data is available
	if( DataType::MapEnum == refreshMsg.getPayload().getDataType() )
		decode( refreshMsg.getPayload().getMap(), false );
}


void AppClient::onUpdateMsg( const UpdateMsg& updateMsg, const OmmConsumerEvent& ) {
	//  display item and service name
	if( updateMsg.hasMsgKey() )
		cout << endl << "Item Name: " << updateMsg.getName() << " (Service Name: " << updateMsg.getServiceName() << ")" << endl;

	// Level2 data payload is encoded as an OMM Map or NoData if no payload data is available
	if (DataType::MapEnum == updateMsg.getPayload().getDataType())
		decode( updateMsg.getPayload().getMap(), true );
}



void AppClient::onStatusMsg( const StatusMsg& statusMsg, const OmmConsumerEvent& )	{
	//  display item and service name
	if( statusMsg.hasMsgKey() )
		cout << endl << "Item Name: " << statusMsg.getName() << endl << "Service Name: " << statusMsg.getServiceName();

	if( statusMsg.hasState() )	{
		//  display the status of current request
		cout << "Stream state: " << statusMsg.getState().getStreamState()
			<< "\tStatus code: " << statusMsg.getState().getStatusCode()
			<< "\tStatus text: " << statusMsg.getState().getStatusText() << endl;
	}
}


void AppClient::decode( const FieldList& fl )	{
	// recurse over all field entries in list
	while( fl.forth() )	{
		const FieldEntry& fe = fl.getEntry();
	
		cout << "\tFid: " << fe.getFieldId() << " Name: " << fe.getName() << " DataType: " << DataType( fe.getLoadType() ).toString() << " Value: ";

		// decode primitive OMM data type of this entry
		if( fe.getCode() == Data::BlankEnum )
			cout << " blank" << endl;
		else	{
			switch ( fe.getLoadType() )		{
				case DataType::RealEnum :
					cout << fe.getReal().getAsDouble() << endl;
					break;
				case DataType::DateEnum :
					cout << (UInt64)fe.getDate().getDay() << " / " << (UInt64)fe.getDate().getMonth() << " / " << (UInt64)fe.getDate().getYear() << endl;
					break;
				case DataType::TimeEnum :
					cout << (UInt64)fe.getTime().getHour() << ":" << (UInt64)fe.getTime().getMinute() << ":" << (UInt64)fe.getTime().getSecond() << ":" << (UInt64)fe.getTime().getMillisecond() << endl;
					break;
				case DataType::IntEnum :
					cout << fe.getInt() << endl;
					break;
				case DataType::UIntEnum :
					cout << fe.getUInt() << endl;
					break;
				case DataType::AsciiEnum :
					cout << fe.getAscii() << endl;
					break;
				case DataType::EnumEnum :
					cout << fe.getEnum() << endl;
					break;
				case DataType::ErrorEnum :
					cout << "( " << fe.getError().getErrorCodeAsString() << " )" << endl;
				default:
					cout << endl;
					break;
			}
		}
	}
}



void AppClient::decode( const Map& map, bool echo )	{
	// if map has summary data, it is of type OMM Fieldlist
	if( map.getSummaryData().getDataType() == DataType::FieldListEnum )	{
		cout << endl << "----------------------------------" << endl;
		cout << "Map Summary:" << endl;
		decode( map.getSummaryData().getFieldList() );
	}

	bool display = true;
	int records = 0;

	// loop over map entries
	while ( map.forth() )	{
		const MapEntry& me = map.getEntry();

		// map entry is a key/data pair, key is an OMM buffer
		if (me.getKey().getDataType() == DataType::BufferEnum) {
			records++;
			if (display) {
				cout << endl << "----------------------------------" << endl;
				const EmaBuffer& keyBuf = me.getKey().getBuffer();
				DataType::DataTypeEnum dType = me.getLoadType();

				// data for Add and Update actions is OMM Fieldlist
				// there is no data for Delete action
				switch (me.getAction()) {

				case MapEntry::AddEnum:
					cout << "ADD Entry" << endl;
					if (dType != DataType::FieldListEnum)
						return;
					decode(me.getFieldList());
					break;

				case MapEntry::DeleteEnum:
					cout << "DELETE Entry" << endl;
					if (dType != DataType::NoDataEnum)
						return;
					break;

				case MapEntry::UpdateEnum:
					cout << "UPDATE Entry" << endl;
					if (dType != DataType::FieldListEnum)
						return;
					decode(me.getFieldList());
					break;
				}
				display = echo;
			}
		}		
	}
	cout << endl << "Total items in this event:" << records << endl;
}	

int main( int argc, char* argv[] )	{
	
	try {
		// instantiate callback client
		AppClient client;
		
		cout << "Connecting to market data server..." << endl;

		// create OMM consumer
		OmmConsumer consumer( 
			OmmConsumerConfig()
				.host( "ADS:14002" )
				.username( "user" ) 
		);
		
		cout << "Subscribing to Level 2 market data" << endl;

		// subscribe to Level2 market data
		consumer.registerClient( 
			ReqMsg()
				.domainType( MMT_MARKET_BY_ORDER )
				.serviceName( "ELEKTRON_DD" )
				.name( "TD.TO" ),
			client );
		
		// block this thread for a while
		sleep(60000);
		
	} 
	catch ( const OmmException& excp ) {
		cout << "Exception subscribing to market data: " << excp << endl;
	}
	
	return 0;
}
